/*

    Asf Sdf Compiler Runtime Library
    Copyright (C) 2000  Stichting Mathematisch Centrum, Amsterdam, 
                        The Netherlands. 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

*/
/**
  * Library support for AsFix2C compiled modules.
  *
  * Mark van den Brand, Pieter Olivier, Jurgen Vinju
  * $Id: asc-support.c,v 1.21 2001/03/08 13:49:55 markvdb Exp $
  */


#include <stdio.h>
#include <assert.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>

#include <aterm2.h>
#include <PT-utils.h>

#include "asc-support.h"

#ifndef streq
#define streq(s,t)    (!(strcmp(s,t)))
#endif


#define MAX_STORE 10240
#define MAGIC_HASH_CONST_APPL 3001
#define MAGIC_HASH_CONST_LIST 4507

#define HASH_PROD(prod, size) (((unsigned)(((int)(prod))>>2)) % (size))
#define HASH_SYM(sym, size) (((unsigned)(((int)(sym))>>2)) % (size))

#define MAX_MEMO_TABLES 1024

/*
 *      MuASF function names. 
 */

typedef struct bucket
{
  struct bucket *next_prod;
  struct bucket *next_sym;
  ATerm prod;
  funcptr func;
  Symbol sym;
} bucket;


/* This term_store is used for temporary storage of
 * list elements, e.g. used in slice.
 */
static ATerm term_store[MAX_STORE];

/* Data structures for storing the MuASF names and the
 * corresponding AsFix names.
 */
static int nr_entries = 0;
static int table_size = 0;
static bucket **prod_table = NULL;
static bucket **sym_table = NULL;

/* Name speaks for itself, a temporary buffer to store
 * strings.
 */
static char conversionbuf[1024];

/* Variables used for profiling. */
unsigned int rewrite_steps = 0;

#ifdef MEMO_PROFILING
ATermTable prof_table = NULL;
Symbol record_sym = -1;
#endif

/* A table to convert characters to ATermInts. 
 */
ATerm char_table[256] = {NULL};

/* These quoted symbols are used in combination with
 * outermost evaluation! So, to enlighten Jurgen
 * "if" Bool "then" S "else" S "fi" -> S {delay2, delay3}
 * Symbols are quoted to prevent innermost reduction of
 * functions. Also symbols related to list (matching)
 * have to be quoted: make_listsym, concsym and conssym.
 */
Symbol sym_quote0;
Symbol sym_quote1;
Symbol sym_quote2;
Symbol sym_quote3;
Symbol sym_quote4;
Symbol sym_quote5;
Symbol sym_quote6;
Symbol sym_quote7;
Symbol make_listsym;

ATerm pattern_listtype;
ATerm pattern_listtype_sep;

/* These symbols are used in combination with lists
 * "conssym" should be removed in the near future! 
 */
Symbol concsym;
Symbol conssym;

/* Declarations of static functions. */
static ATermList innermost_list(PT_Args args);
static funcptr basic_lookup_func(ATerm prod);
static ATerm call(PT_Production prod, ATermList args);
static ATerm make_list_type(PT_Symbol type, ATermList args);
static ATermList innermost_list(PT_Args args);
static ATerm innermost_lexical(PT_Tree lex);
static PT_Tree term_to_asfix(ATerm t, PT_Symbol sort);
static PT_Tree make_asfix_list(ATermList l, char *sort);
static PT_Tree make_asfix_list_sep(ATermList l, char *sort, char *sep);
static int get_list_length(ATermList chars);
static PT_Args terms_to_asfix(PT_Symbols a, ATermAppl t, PT_Symbol sort);
static ATbool isLexicalConstructorProd(PT_Production prod);
static void stringToLower(char *str);

/* Variables needed when counting memo table size */
static int nr_memo_tables = 0;
static ATermTable *memo_tables[MAX_MEMO_TABLES];
static char *memo_table_names[MAX_MEMO_TABLES];

/**
  * Create a new (larger?) hash table for register and 
  * resolve of context-free function names and rehash all 
  * old the entries.
  */

void c_rehash(int newsize)
{
  int i; 
  unsigned int hnr;
  bucket **newprodtable, **newsymtable, *b, *next;

  /* Allocate new hash tables */
  newprodtable = (bucket **)calloc(newsize, sizeof(bucket *));
  newsymtable = (bucket **)calloc(newsize, sizeof(bucket *));
  if(!newsymtable || !newprodtable) {
    ATabort("out of memory in c_rehash(%d)", newsize);
  }

  /* Rehash all the old buckets. */
  if(prod_table) {
    for(i=0; i<table_size; i++) {
      b = prod_table[i];
      while(b) {
	next = b->next_prod;
	hnr = HASH_PROD(b->prod, newsize);
	b->next_prod = newprodtable[hnr];
	newprodtable[hnr] = b;
	b = next;
      }
      b = sym_table[i];
      while(b) {
	next = b->next_sym;
	hnr = HASH_SYM(b->sym, newsize);
	b->next_sym = newsymtable[hnr];
	newsymtable[hnr] = b;
	b = next;
      }
    }
  }

  if(prod_table) {
    free(prod_table);
    free(sym_table);
  }
  prod_table = newprodtable;
  sym_table = newsymtable;
  table_size = newsize;
}



#ifdef NO_SHARING
/*
 * Calculate hash function of a term.
 */

unsigned int calc_hash(ATerm t)
{
  unsigned int hnr = 0;

  switch(ATgetType(t)) {
  case AT_APPL:
    {
      ATermAppl appl = (ATermAppl)t;
      AFun sym = ATgetAFun(appl);
      int i, arity = ATgetArity(sym);
      hnr = AT_hashSymbol(ATgetName(sym), arity);

      for(i=0; i<arity; i++) {
				hnr = hnr * MAGIC_HASH_CONST_APPL + 
					calc_hash(ATgetArgument(appl, i));
      }
    }
    break;

  case AT_INT:
    hnr = ATgetInt((ATermInt)t);
    break;

  case AT_LIST:
    {
      ATermList list = (ATermList)t;
      hnr = 123;
      while(!ATisEmpty(list)) {
				hnr = hnr * MAGIC_HASH_CONST_LIST + 
					calc_hash(ATgetFirst(list));
				list = ATgetNext(list);
      }
    }
    break;
  }

  return hnr;
}
#endif



void register_prod(ATerm prod, funcptr func, Symbol sym)
{
  unsigned int hnr; /* hash number */
  bucket *b; /* single bucket */

  /* Heuristics for deciding when to rehash */
  if((nr_entries*100)/table_size > MAX_LOAD) {
    c_rehash(table_size*2);
  }

#ifdef NO_SHARING
  /* calc_hash is recursive, so no mod included */
  hnr = calc_hash(prod); 
  hnr %= table_size;
#else
  hnr = HASH_PROD(prod, table_size);
#endif

  /* Find out if this function has already been registered */
  b = prod_table[hnr];
  while(b && !ATisEqual(b->prod, prod))
    b = b->next_prod;

  /* if already registered, we are done */
  if(b)
    return;

  /* Else we should add a new bucket */
  b = malloc(sizeof(bucket));
  if(!b)
    ATabort("out of memory in register.\n");
  b->next_prod = prod_table[hnr];
  prod_table[hnr] = b;

  /* Add it to the sym table also. */
  hnr = HASH_SYM(sym, table_size);
  b->next_sym = sym_table[hnr];
  sym_table[hnr] = b;

  /* The prod field of the bucket is protected against 
   * the gc.
   */
  b->prod = prod;
  ATprotect((ATerm*)&(b->prod));
  b->func = func;
  b->sym  = sym;
}

/* The following functions are used to convert from symbols
 * to function pointers or from prods to function pointers.
 * And from symbols to prods. This functionality is used when
 * converting back a normal form to AsFix.
 */


funcptr lookup_func_given_sym(Symbol sym)
{
  bucket *b;
  unsigned int hnr;

  hnr = HASH_SYM(sym, table_size);

  b = sym_table[hnr];

  while(b) {
    if(b->sym == sym)
      return b->func;
    b = b->next_sym;
  }
  ATabort("unknown symbol: %s\n", ATgetName(sym));
  return (funcptr) NULL; /* silence the compiler, we never get here. */
}

/* Changed this function to static 15-02-2000, await regression test */
static funcptr basic_lookup_func(ATerm prod)
{
  bucket *b;
  unsigned int hnr;

#ifdef NO_SHARING
  hnr = calc_hash(prod);
  hnr %= table_size;
#else
  hnr = HASH_PROD(prod, table_size);
#endif
  
  b = prod_table[hnr];

  while(b) {
    if(ATisEqual(b->prod, prod))
      return b->func;
    b = b->next_prod;
  } 

  return (funcptr) NULL;
}

funcptr lookup_func(ATerm prod)
{
  funcptr f = basic_lookup_func(prod);
  if(!f) {
    ATabort("unknown function: %t\n", prod);
    return (funcptr) NULL; /* silence the compiler, we never get here. */
  }
  else return f;
}

Symbol lookup_sym(ATerm prod)
{
  bucket *b;
  unsigned int hnr;

#ifdef NO_SHARING
  hnr = calc_hash(prod);
  hnr %= table_size;
#else
  hnr = HASH_PROD(prod, table_size);
#endif
  
  b = prod_table[hnr];

  while(b) {
    if(ATisEqual(b->prod, prod))
      return b->sym;
    b = b->next_prod;
  }
  ATabort("Unknown function: %t\n", prod);
  return (Symbol) 0; /* silence the compiler, we never get here. */
}


ATerm lookup_prod(Symbol sym)
{
  bucket *b;
  unsigned int hnr;

  hnr = HASH_SYM(sym, table_size);

  b = sym_table[hnr];

  while(b) {
    if(b->sym == sym)
      return b->prod;
    b = b->next_sym;
  }
  ATabort("unknown symbol: %s\n", ATgetName(sym));
  return (ATerm) NULL; /* silence the compiler, we never get here. */
}

/**
  * Translate a string to a list of integers (characters)
  */

ATermList string2list(char *s)
{
  int i;
  ATermList result = ATempty;

  for(i=strlen(s)-1; i>=0; i--) {
    result = ATinsert(result, make_char((int)((unsigned char)s[i])));
  }

  return result;
}

/* This function is used in innermost to call a c function for
 * a given AsFix production. Note that the arguments are in normal form
 * because of the innermost reduction strategy. Apparently there is
 * an upper limit on the number of arguments of a production (32)!
 *
 * The functionality of this function is to convert the argument list in
 * ATermList form to a c function call and to lookup the actual funcptr.
 */

static ATerm call(PT_Production prod, ATermList args)
{
  ATermList list;
  ATerm arg[33];
  int idx = 0;
  funcptr func = basic_lookup_func(PT_makeTermFromProduction(prod));

  if (func == NULL) {
    ATabort("unknown function: %t\n", prod);
    return (ATerm) NULL; 
  }

  list = args;
  while(!ATisEmpty(list)) {
    arg[idx++] = ATgetFirst(list);
    list = ATgetNext(list);
  }

  /* We cannot use C's variable argument list support,
     because there is no portable way to build a va_list 
     from an arbitrary list.
   */
  switch(ATgetLength(args)) {
  case  0: return (*func)();
  case  1: return (*func)(arg[0]);
  case  2: return (*func)(arg[0],arg[1]);
  case  3: return (*func)(arg[0],arg[1],arg[2]);
  case  4: return (*func)(arg[0],arg[1],arg[2],arg[3]);
  case  5: return (*func)(arg[0],arg[1],arg[2],arg[3],arg[4]);
  case  6: return (*func)(arg[0],arg[1],arg[2],arg[3],arg[4],arg[5]);
  case  7: return (*func)(arg[0],arg[1],arg[2],arg[3],arg[4],arg[5],
			  arg[6]);
  case  8: return (*func)(arg[0],arg[1],arg[2],arg[3],arg[4],arg[5],
			  arg[6],arg[7]);
  case  9: return (*func)(arg[0],arg[1],arg[2],arg[3],arg[4],arg[5],
			  arg[6],arg[7],arg[8]);
  case 10: return (*func)(arg[0],arg[1],arg[2],arg[3],arg[4],arg[5],
			  arg[6],arg[7],arg[8],arg[9]);
  case 11: return (*func)(arg[0],arg[1],arg[2],arg[3],arg[4],arg[5],
			  arg[6],arg[7],arg[8],arg[9],arg[10]);
  case 12: return (*func)(arg[0],arg[1],arg[2],arg[3],arg[4],arg[5],
			  arg[6],arg[7],arg[8],arg[9],arg[10],arg[11]);
  case 13: return (*func)(arg[0],arg[1],arg[2],arg[3],arg[4],arg[5],
			  arg[6],arg[7],arg[8],arg[9],arg[10],arg[11],
			  arg[12]);
  case 14: return (*func)(arg[0],arg[1],arg[2],arg[3],arg[4],arg[5],
			  arg[6],arg[7],arg[8],arg[9],arg[10],arg[11],
			  arg[12],arg[13]);
  case 15: return (*func)(arg[0],arg[1],arg[2],arg[3],arg[4],arg[5],
			  arg[6],arg[7],arg[8],arg[9],arg[10],arg[11],
			  arg[12],arg[13],arg[14]);
  case 16: return (*func)(arg[0],arg[1],arg[2],arg[3],arg[4],arg[5],
			  arg[6],arg[7],arg[8],arg[9],arg[10],arg[11],
			  arg[12],arg[13],arg[14],arg[15]);
  case 17: return (*func)(arg[0],arg[1],arg[2],arg[3],arg[4],arg[5],
			  arg[6],arg[7],arg[8],arg[9],arg[10],arg[11],
			  arg[12],arg[13],arg[14],arg[15],arg[16]);
  case 18: return (*func)(arg[0],arg[1],arg[2],arg[3],arg[4],arg[5],
			  arg[6],arg[7],arg[8],arg[9],arg[10],arg[11],
			  arg[12],arg[13],arg[14],arg[15],arg[16],arg[17]);
  case 19: return (*func)(arg[0],arg[1],arg[2],arg[3],arg[4],arg[5],
			  arg[6],arg[7],arg[8],arg[9],arg[10],arg[11],
			  arg[12],arg[13],arg[14],arg[15],arg[16],arg[17],
			  arg[18]);
  case 20: return (*func)(arg[0],arg[1],arg[2],arg[3],arg[4],arg[5],
			  arg[6],arg[7],arg[8],arg[9],arg[10],arg[11],
			  arg[12],arg[13],arg[14],arg[15],arg[16],arg[17],
			  arg[18],arg[19]);
  case 21: return (*func)(arg[0],arg[1],arg[2],arg[3],arg[4],arg[5],
			  arg[6],arg[7],arg[8],arg[9],arg[10],arg[11],
			  arg[12],arg[13],arg[14],arg[15],arg[16],arg[17],
			  arg[18],arg[19],arg[20]);
  case 22: return (*func)(arg[0],arg[1],arg[2],arg[3],arg[4],arg[5],
			  arg[6],arg[7],arg[8],arg[9],arg[10],arg[11],
			  arg[12],arg[13],arg[14],arg[15],arg[16],arg[17],
			  arg[18],arg[19],arg[20],arg[21]);
  case 23: return (*func)(arg[0],arg[1],arg[2],arg[3],arg[4],arg[5],
			  arg[6],arg[7],arg[8],arg[9],arg[10],arg[11],
			  arg[12],arg[13],arg[14],arg[15],arg[16],arg[17],
			  arg[18],arg[19],arg[20],arg[21],arg[22]);
  case 24: return (*func)(arg[0],arg[1],arg[2],arg[3],arg[4],arg[5],
			  arg[6],arg[7],arg[8],arg[9],arg[10],arg[11],
			  arg[12],arg[13],arg[14],arg[15],arg[16],arg[17],
			  arg[18],arg[19],arg[20],arg[21],arg[22],arg[23]);
  case 25: return (*func)(arg[0],arg[1],arg[2],arg[3],arg[4],arg[5],
			  arg[6],arg[7],arg[8],arg[9],arg[10],arg[11],
			  arg[12],arg[13],arg[14],arg[15],arg[16],arg[17],
			  arg[18],arg[19],arg[20],arg[21],arg[22],arg[23],
			  arg[24]);
  case 26: return (*func)(arg[0],arg[1],arg[2],arg[3],arg[4],arg[5],
			  arg[6],arg[7],arg[8],arg[9],arg[10],arg[11],
			  arg[12],arg[13],arg[14],arg[15],arg[16],arg[17],
			  arg[18],arg[19],arg[20],arg[21],arg[22],arg[23],
			  arg[24],arg[25]);
  case 27: return (*func)(arg[0],arg[1],arg[2],arg[3],arg[4],arg[5],
			  arg[6],arg[7],arg[8],arg[9],arg[10],arg[11],
			  arg[12],arg[13],arg[14],arg[15],arg[16],arg[17],
			  arg[18],arg[19],arg[20],arg[21],arg[22],arg[23],
			  arg[24],arg[25],arg[26]);
  case 28: return (*func)(arg[0],arg[1],arg[2],arg[3],arg[4],arg[5],
			  arg[6],arg[7],arg[8],arg[9],arg[10],arg[11],
			  arg[12],arg[13],arg[14],arg[15],arg[16],arg[17],
			  arg[18],arg[19],arg[20],arg[21],arg[22],arg[23],
			  arg[24],arg[25],arg[26],arg[27]);
  case 29: return (*func)(arg[0],arg[1],arg[2],arg[3],arg[4],arg[5],
			  arg[6],arg[7],arg[8],arg[9],arg[10],arg[11],
			  arg[12],arg[13],arg[14],arg[15],arg[16],arg[17],
			  arg[18],arg[19],arg[20],arg[21],arg[22],arg[23],
			  arg[24],arg[25],arg[26],arg[27],arg[28]);
  case 30: return (*func)(arg[0],arg[1],arg[2],arg[3],arg[4],arg[5],
			  arg[6],arg[7],arg[8],arg[9],arg[10],arg[11],
			  arg[12],arg[13],arg[14],arg[15],arg[16],arg[17],
			  arg[18],arg[19],arg[20],arg[21],arg[22],arg[23],
			  arg[24],arg[25],arg[26],arg[27],arg[28],arg[29]);
  case 31: return (*func)(arg[0],arg[1],arg[2],arg[3],arg[4],arg[5],
			  arg[6],arg[7],arg[8],arg[9],arg[10],arg[11]
			  ,arg[12],arg[13],arg[14],arg[15],arg[16],arg[17],
			  arg[18],arg[19],arg[20],arg[21],arg[22],arg[23],
			  arg[24],arg[25],arg[26],arg[27],arg[28],arg[29],
			  arg[30]);
  case 32: return (*func)(arg[0],arg[1],arg[2],arg[3],arg[4],arg[5],
			  arg[6],arg[7],arg[8],arg[9],arg[10],arg[11],
			  arg[12],arg[13],arg[14],arg[15],arg[16],arg[17],
			  arg[18],arg[19],arg[20],arg[21],arg[22],arg[23],
			  arg[24],arg[25],arg[26],arg[27],arg[28],arg[29],
			  arg[30],arg[31]);
  case 33: return (*func)(arg[0],arg[1],arg[2],arg[3],arg[4],arg[5],
			  arg[6],arg[7],arg[8],arg[9],arg[10],arg[11],
			  arg[12],arg[13],arg[14],arg[15],arg[16],arg[17],
			  arg[18],arg[19],arg[20],arg[21],arg[22],arg[23],
			  arg[24],arg[25],arg[26],arg[27],arg[28],arg[29],
			  arg[30],arg[31],arg[32]);
  default:
    ATabort("too many arguments: %d\n", ATgetLength(args));
  }
  return NULL;
}


/* This function is also used by innermost. It maps the AsFix representation
 * of a list to the C runtime representation of lists. The arguments are
 * already in normal form due to the innermost reduction strategy. A list
 * can never occur as an outermost function symbol (it's a builtin), so
 * no 'call' is needed.
 * 
 */

static ATerm make_list_type(PT_Symbol type, ATermList args)
{
  char *sort, *sep;
  char symbol_buf[256];
  Symbol sym;

  /* The list has no separator. */
  if (PT_isSymbolIterStar(type) || PT_isSymbolIterPlus(type)) {
    sort = strdup(PT_getSymbolString(PT_getSymbolSymbol(type)));
    sprintf(symbol_buf, "listtype(sort(\"%s\"))", sort);
    sym = ATmakeSymbol(symbol_buf, 1, ATtrue);
    free(sort); 
    return (ATerm) ATmakeApplList(sym, ATmakeList1((ATerm)args));

  } /* Or the list has a separator defined. */
  else if(PT_isSymbolIterStarSep(type) || PT_isSymbolIterStarSep(type)) {
    sort = strdup(PT_getSymbolString(PT_getSymbolSymbol(type)));
    sep  = PT_getSymbolSeparator(type);
    sprintf(symbol_buf, "listtype(sort(\"%s\"),ql(\"%s\"))", sort, sep);
    sym = ATmakeSymbol(symbol_buf, 1, ATtrue);
    free(sort); 
    return (ATerm) ATmakeApplList(sym, ATmakeList1((ATerm)args));
  }
  else {
    ATabort("illegal list type: %t\n", type);
    return (ATerm) NULL; /* silence the compiler */
  }
}


void reg_memo_table(ATermTable *db, char *name)
{
  memo_tables[nr_memo_tables] = db;
  memo_table_names[nr_memo_tables] = name;

  nr_memo_tables++;
}


void print_memo_table_sizes()
{
  int i, nr_key_nodes, nr_value_nodes;
  ATermList keys, values;

  for (i=0; i<nr_memo_tables; i++) {
    keys = ATtableKeys(*memo_tables[i]);
    values = ATtableValues(*memo_tables[i]);
    nr_key_nodes = ATcalcUniqueSubterms((ATerm) keys);
    nr_value_nodes = ATcalcUniqueSubterms((ATerm) values);

    fprintf(stderr, "table %s has %d elements, %d + %d = %d unique nodes.\n",
	    memo_table_names[i], ATgetLength(keys), 
	    nr_key_nodes, nr_value_nodes, nr_key_nodes+nr_value_nodes);
  }
}


static ATerm innermost_lexical(PT_Tree lex)
{
  PT_Symbol sort;
  PT_Production prod;
  char *name;
  ATermList lexlist;
  ATerm listterm;
 
  sort = PT_getTreeSymbol(lex);
  
  name = strdup(PT_getSymbolString(sort));
  if (name == NULL) {
    ATerror("innermost_lexical: could not allocate memory for sortname\n");
  }

  stringToLower(name);

  lexlist = string2list(PT_getTreeString(lex));
   
  prod = (PT_Production) ASF_makeProductionLexicalConstructor(name, 
			   (ASF_Symbol) sort);

  free(name);
  
  listterm = ATmake("\"listtype(sort(\\\"CHAR\\\"))\"([<list>])", lexlist);

  return call(prod, ATmakeList1(listterm));
}

/* This function is the core of the innermost reduction strategy. Given the
 * AsFIx representation of a term to be reduced the c functions for rewriting
 * are called in an innermost manner.  Here we use the "call-by-value" mechanism
 * of C to implement the innermost strategy.
 */

ATerm innermost(PT_Tree tree)
{
  ATerm result = NULL;

  if (PT_isTreeAppl(tree)) {
    PT_Production prod = PT_getTreeProd(tree);
    PT_Args args = PT_getTreeArgs(tree);

    if (PT_hasProductionBracketAttr(prod)) {
      result = innermost(PT_getArgsArgumentAt(args, 2));    
    } 
    else {
      result = call(prod, innermost_list(args));
    }
  }
  else if (PT_isTreeLexical(tree)) {
    result = innermost_lexical(tree);
  }
  else if (PT_isTreeList(tree)) {
    PT_Args args = PT_getTreeArgs(tree);
    result = make_list_type(PT_getTreeIter(tree), innermost_list(args));
  }

  return result;
}


/* The function innermost_list takes care of the normalization
 * of the elements of a list (of arguments) with respect to the innermost
 * reduction strategy. Here we use the "call-by-value" mechanism
 * of C to implement the innermost strategy.
 */

static ATermList innermost_list(PT_Args args)
{
  ATermList result = ATempty;
  int length = PT_getArgsLength(args);
  ATerm el;
 
  /* When a list is shorter than 16 elements, we assume that
     it is not worth it to malloc a buffer. We use ATreverse
     to restore the correct list order. */
  if(length < 16) {
    while(PT_hasArgsHead(args)) {
      el = innermost(PT_getArgsHead(args));
      if(el)  
	result = ATinsert(result, el); 
      args = PT_getArgsTail(args);
    }
    return ATreverse(result);
  } else {

    /* We don't use term_store here because of the
     * recursive calls to innermost, and the fact that
     * slice (which uses term_store) may be called by
     * the generated code.
     */
    int idx = 0;
    ATerm *elems = (ATerm *)malloc(sizeof(ATerm)*length);
    if(!elems) {
      ATabort("innermost_list: no room for %d elements.\n", length);
    }

    while(PT_hasArgsHead(args)) {
      elems[idx++] = PT_makeTermFromTree(PT_getArgsHead(args));
      args = PT_getArgsTail(args);
    }
    assert(idx == length);

    for(--idx; idx>=0; idx--) {
      el = innermost(PT_makeTreeFromTerm(elems[idx]));
      if(el) {
	result = ATinsert(result, el);
      }
    }
		
    free(elems);
    return result;
  }
}



/* Code to unqoute delayed reduction of terms, in order to implement
 * the outermost strategy. This algorithm is known to be 'not optimal'.
 * We refer to the work of Wan Fokkink, Pum Walters and Jasper Kamperman
 * for better implementation ideas.
 */

ATerm unquote(ATerm t)
{
  Symbol s;
  ATerm a0,a1,a2,a3,a4,a5,a6;

  if (ATmatch(t,"quote(<int>)",&s)) {
    funcptr func = lookup_func_given_sym(s);
    return (*func)();
  }
  if (ATmatch(t,"quote(<int>,<term>)",&s,&a0)) {
    if (s == make_listsym)
      return make_list(unquote(a0));
    else {
      funcptr func = lookup_func_given_sym(s);
      return (*func)(unquote(a0));
    }
  }
  if (ATmatch(t,"quote(<int>,<term>,<term>)",&s,&a0,&a1)) {
    if (s == concsym || s == conssym)
      return (ATerm)ATconcat((ATermList)unquote(a0),(ATermList)unquote(a1));
    else {
      funcptr func = lookup_func_given_sym(s);
      return (*func)(unquote(a0),unquote(a1));
    }
  }
  if (ATmatch(t,"quote(<int>,<term>,<term>,<term>)",&s,&a0,&a1,&a2)) {
    funcptr func = lookup_func_given_sym(s);
    return (*func)(unquote(a0),unquote(a1),unquote(a2));
  }
  if (ATmatch(t,"quote(<int>,<term>,<term>,<term>,<term>)",
              &s,&a0,&a1,&a2,&a3)) {
    funcptr func = lookup_func_given_sym(s);
    return (*func)(unquote(a0),unquote(a1),unquote(a2),unquote(a3));
  }
  if (ATmatch(t,"quote(<int>,<term>,<term>,<term>,<term>,<term>)",
              &s,&a0,&a1,&a2,&a3,&a4)) {
    funcptr func = lookup_func_given_sym(s);
    return (*func)(unquote(a0),unquote(a1),unquote(a2),unquote(a3),unquote(a4));
  }
  if (ATmatch(t,"quote(<int>,<term>,<term>,<term>,<term>,<term>,<term>)",
              &s,&a0,&a1,&a2,&a3,&a4,&a5)) {
    funcptr func = lookup_func_given_sym(s);
    return (*func)(unquote(a0),unquote(a1),unquote(a2),unquote(a3),unquote(a4),unquote(a5));
  }
  if (ATmatch(t,"quote(<int>,<term>,<term>,<term>,<term>,<term>,<term>,<term>)",
              &s,&a0,&a1,&a2,&a3,&a4,&a5,&a6)) {
    funcptr func = lookup_func_given_sym(s);
    return (*func)(unquote(a0),unquote(a1),unquote(a2),unquote(a3),unquote(a4),unquote(a5),unquote(a6));
  }

  return t;
}


/* The following set of functions take care of the "translation" back from
 * the C/MuASF representation of terms to the AsFix representation.
 * This conversion back is only needed when the normal form is obtained.
 */


/* A list is converted back into an AsFix list. 
   The elements of the list are recursively converted to AsFix.
   Again, term_store cannot be used because of the
   recursive nature of this function.
 */

static PT_Tree make_asfix_list(ATermList l, char *sort)
{
  PT_Symbol listsort = PT_makeSymbolSort(sort);
  PT_Symbol iter = PT_makeSymbolIterStar(listsort);
  PT_Args list = PT_makeArgsEmpty();
  PT_Tree tmp;
  PT_Tree newline = PT_makeTreeLayout("\n");

  int i, len = ATgetLength(l);
  if(len > 0) {
    tmp = term_to_asfix(ATelementAt(l, len-1), listsort);
    list = PT_makeArgsList(tmp,list);
    for(i = len-2; i >= 0; i--) {
      list = PT_makeArgsList(newline, list);
      tmp = term_to_asfix(ATelementAt(l, i), listsort);
      list = PT_makeArgsList(tmp,list);
    }
  }
  return PT_makeTreeList(iter, list);  
}


/* The same as the previous function, but now with a separated list.
 */

static PT_Tree make_asfix_list_sep(ATermList l, char *sort, char *sep)
{
  PT_Args list = PT_makeArgsEmpty();
  PT_Tree tmp;
  int i;

  PT_Symbol listsort = PT_makeSymbolSort(sort);
  PT_Symbol iter = PT_makeSymbolIterStarSep(listsort,sep);
  PT_Tree   tsep = PT_makeTreeSeparator(sep);
  PT_Tree   space = PT_makeTreeLayout(" ");
  PT_Tree   newline = PT_makeTreeLayout("\n");

  int len = ATgetLength(l);
  if(len > 0) {
    tmp = term_to_asfix(ATelementAt(l, len-1), listsort);
    list = PT_makeArgsList(tmp, list);
    for(i = len-2; i >= 0; i--) {
      list = PT_makeArgsList(space, list);
      list = PT_makeArgsList(tsep, list);
      list = PT_makeArgsList(newline, list);
      tmp = term_to_asfix(ATelementAt(l, i), listsort);
      list = PT_makeArgsList(tmp,list);
    }
  }

  return PT_makeTreeList(iter, list);  
}


/*
 * Calculates the length of a list of characters. 
 * Pieter? According to Jurgen and me this code can
 * be tremendously simplified. We assume that character
 * lists have a nested structure.
 */

static int get_list_length(ATermList chars)
{
  return ATgetLength(chars);
}


static char *get_chars_from_list(char *buf, ATermList chars)
{
  while(!ATisEmpty(chars)) {
    ATerm el = ATgetFirst(chars);
    chars = ATgetNext(chars);
    *buf++ = (char)ATgetInt((ATermInt) el);
  }
  return buf;
}

static ATbool isLexicalConstructorProd(PT_Production prod)
{
  ASF_Production lexicalProd = ASF_makeProductionFromTerm(
			        PT_makeTermFromProduction(prod));

  if (ASF_isValidProduction(lexicalProd)) {
    return ASF_isProductionLexicalConstructor(lexicalProd);
  }

  return ATfalse;  
}

static PT_Tree make_asfix_lex(ATermList chars, PT_Symbol sort)
{
  char *buf, *end;
  PT_Tree result;
  int len = get_list_length(chars);

  buf = malloc(len+1);
  if(!buf) {
    ATabort("out of memory trying to allocate lex of %d chars\n", len);
    return NULL;
  }
  end = get_chars_from_list(buf, chars);
  *end = '\0';
  result = PT_makeTreeLexical(buf, sort);
  free(buf);
  return result;
}



/* The top function for converting back
 */

PT_ParseTree toasfix(ATerm term)
{ 
  return PT_makeParseTreeTree("",term_to_asfix(term,NULL),"");
}



static PT_Tree term_to_asfix(ATerm tree, PT_Symbol sort)
{
  PT_Tree result = NULL;
  PT_Symbol lexsort;
  ATerm prod;
  PT_Production ptProd;
  PT_Symbols formalargs;
  PT_Args actualargs;
  char *listsort, *sep;

  if(ATgetType(tree) == AT_APPL) {
    prod = lookup_prod(ATgetSymbol((ATermAppl)tree));
    if(!prod) {
      ATabort("unknown production symbol: %s\n",
              ATgetName(ATgetSymbol((ATermAppl) tree)));
    }

    ptProd = PT_makeProductionFromTerm(prod);
    if (PT_isValidProduction(ptProd)) {
      if (isLexicalConstructorProd(ptProd)) {
        lexsort = PT_getProductionRhs(ptProd);
        return term_to_asfix(ATgetArgument((ATermAppl) tree, 0), lexsort);
      } 
      else {
      /* An ordinary production rule is processed. 
       */
	formalargs = PT_getProductionLhs(ptProd);
	sort = PT_getProductionRhs(ptProd);
	actualargs = terms_to_asfix(formalargs, (ATermAppl)tree, sort);

	result = PT_makeTreeAppl(ptProd,actualargs);
      }
    }
    else if(ATmatchTerm(prod, pattern_listtype, &listsort)) {
      if(streq(listsort, "CHAR")) {
        result = make_asfix_lex((ATermList)ATgetArgument((ATermAppl) tree,0),
                                sort);
      }
      else {
        result = make_asfix_list((ATermList)ATgetArgument((ATermAppl) tree,0),
                                 listsort);
      }
    }
    /* Conversion of list with separators to AsFix.
     */
    else if(ATmatchTerm(prod, pattern_listtype_sep, &listsort, &sep)) {
      result = make_asfix_list_sep((ATermList) 
                                     ATgetArgument((ATermAppl) tree,0),
                                   listsort, sep);
    }
    else {
      ATabort("cannot handle production: %t\n", prod);
    }
  }
  else {
    ATabort("cannot handle term type: %d\n", ATgetType(tree));
  }

  return result;
}


/*
 * Remove one level of escaping from a string.
 */

void deslash(char *str, char *buf)
{
  while(*str) {
    switch(*str) {
    case '\\':
      str++;
      switch(*str) {
      case '\\':
	break;
      default:
	*buf++ = '\\';
      }
      *buf++ = *str;
      break;
    default:
      *buf++ = *str;
    }
    str++;
  }
  *buf++ = '\0';
}

static void stringToLower(char *str)
{
  int len = strlen(str);

  while(--len >= 0) {
    str[len] = tolower(str[len]);
  }
}

/*
 * Translate a list of C/MuASF terms back to AsFix. Starting at the end of
 * the list of terms, each term is transformed into an AsFix term. There
 * are four different alternatives:
 * 1. It is a whitespace.
 * 2. It is a quoted literal, in that case extra backslashes should be removed.
 * 3. It is a literal, which can be inserted without further processing.
 * 4. It is an application, in this case the sort is retrieved from the 
 *    the list of formal arguments and given this sort the term is transformed
 *    into an AsFix term.
 */

static PT_Args terms_to_asfix(PT_Symbols args, ATermAppl appl, PT_Symbol sort) 
{
  PT_Args result = PT_makeArgsEmpty();
  int arity = ATgetArity(ATgetSymbol(appl));
  int i, j = arity-1, len = PT_getSymbolsLength(args);
  
  for(i=len-1; i>=0; i--) {
    PT_Symbol tmp = PT_getSymbolsSymbolAt(args,i);

    if (PT_isSymbolEmptyLayout(tmp)) {
        result = PT_makeArgsList(PT_makeTreeLayout(" "), result);
    }
    else if(PT_isSymbolQuotedLiteral(tmp)) {
      char *str = strdup(PT_getSymbolString(tmp));
      deslash(str, conversionbuf);
      result = PT_makeArgsList(
	         PT_makeTreeUnquotedLiteral(conversionbuf), result);
      free(str);
    } 
    else if(PT_isSymbolUnquotedLiteral(tmp)) {
      char *str = strdup(PT_getSymbolString(tmp));
      result = PT_makeArgsList(PT_makeTreeUnquotedLiteral(str),result);
      free(str);
    }
    else { 
      PT_Tree tmp2;
      ATerm tmp3 = ATgetArgument(appl, j);
      tmp2 = term_to_asfix(tmp3, sort);
      result = PT_makeArgsList(tmp2,result);
      j--;
    }
  }

  return result;
}

/*
 * Write memo profiling info to the file "memo.prof"
 */

#ifdef MEMO_PROFILING
void write_memo_profile()
{
  FILE *f = fopen("memo.prof", "w");
  ATermList keys;

  if(!f)
    ATabort("cannot create file 'memo.prof'\n");

  keys = ATtableKeys(prof_table);

  while(!ATisEmpty(keys)) {
    ATerm key = ATgetFirst(keys);
    ATermAppl stats = (ATermAppl)ATtableGet(prof_table, key);
    ATerm asfix = lookup_prod(ATgetSymbol((ATermAppl)key));
		
    ATfprintf(f, "%t ", stats);
    AFsourceToFile(asfix, f);
    ATfprintf(f, "\n");
    keys = ATgetNext(keys);
  }
  ATfprintf(f, "\n");

  fclose(f);
}
#endif


/* Part of the initialization code in this function should
 * move to the AsFix library. Constants used during rewriting
 * are initialized here.
 */

void ASC_initRunTime(int tableSize)
{
  int i;

  /* make characters */
  for(i=0; i<256; i++) {
    ATprotect(&char_table[i]);
    char_table[i] = (ATerm) ATmakeInt(i);
  }

  ATprotect(&pattern_listtype);
  ATprotect(&pattern_listtype_sep);

  sym_quote0 = ATmakeSymbol("quote", 1, ATfalse);
  ATprotectSymbol(sym_quote0);
  sym_quote1 = ATmakeSymbol("quote", 2, ATfalse);
  ATprotectSymbol(sym_quote1);
  sym_quote2 = ATmakeSymbol("quote", 3, ATfalse);
  ATprotectSymbol(sym_quote2);
  sym_quote3 = ATmakeSymbol("quote", 4, ATfalse);
  ATprotectSymbol(sym_quote3);
  sym_quote4 = ATmakeSymbol("quote", 5, ATfalse);
  ATprotectSymbol(sym_quote4);
  sym_quote5 = ATmakeSymbol("quote", 6, ATfalse);
  ATprotectSymbol(sym_quote5);
  sym_quote6 = ATmakeSymbol("quote", 7, ATfalse);
  ATprotectSymbol(sym_quote6);
  sym_quote7 = ATmakeSymbol("quote", 8, ATfalse);
  ATprotectSymbol(sym_quote7);
  make_listsym = ATmakeSymbol("make_list", 1, ATfalse);
  ATprotectSymbol(make_listsym);
  concsym = ATmakeSymbol("conc", 2, ATfalse);
  ATprotectSymbol(concsym);
  conssym = ATmakeSymbol("cons", 2, ATfalse);
  ATprotectSymbol(conssym);

  /* muASF list patterns */
  pattern_listtype   = ATparse("listtype(sort(<str>))");
  pattern_listtype_sep = ATparse("listtype(sort(<str>),ql(<str>))");

#ifdef MEMO_PROFILING
  prof_table = ATtableCreate(2048, 80);
  record_sym = ATmakeSymbol("stats", 2, ATfalse);
  ATprotectSymbol(record_sym);

  atexit(write_memo_profile);
#endif

  PT_initPTApi();
  ASF_initASFApi();

  c_rehash(tableSize);
}


/* Constructs the slice of a list starting at index "l1" and ending JUST BEFORE 
 * index "l2". If the slice fits into the term_store a fast way of
 * copying can be used, otherwise the term_store is filled and the
 * rest is processed in a less efficient way.
 */

ATerm slice(ATerm l1, ATerm l2)
{
  ATermList result;
  int i, len;

  if(ATisEmpty((ATermList)l2)) {
    return l1;
  }

  len = slice_length(l1, l2);
  if( MAX_STORE < len )
    len = MAX_STORE;

  for(i=0; i<len; i++) {
    term_store[i] = ATgetFirst((ATermList)l1);
    l1 = (ATerm)ATgetNext((ATermList)l1);
  }

  result = ATempty;

  while(l1 != l2) {
    result = ATappend(result, ATgetFirst((ATermList)l1));
    l1 = (ATerm)ATgetNext((ATermList)l1); 
  }

  for(i=len-1; i>=0; i--) {
    result = ATinsert(result, term_store[i]);
  }
 
  return (ATerm)result;
}


/* Transforms a term into a list. If the term is already a list this
 * list is returned.
 */

ATerm make_list(ATerm t)
{
  if(ATgetType(t) == AT_LIST)  {
    return t;
  }
  else {
    return singleton(t);
  }
}
